@@ -5,18 +5,17 @@
 
 use std::str::FromStr;
 
-use auto_enums::auto_enum;
 use pyo3::{
-    exceptions::{PyException, PyValueError},
+    exceptions::{PyException, PyTypeError, PyValueError},
     ffi,
     prelude::*,
     pybacked::PyBackedStr,
 };
 use text_splitter::{
-    Characters, ChunkCapacity, ChunkCapacityError, ChunkConfig, ChunkConfigError, ChunkSize,
-    ChunkSizer, CodeSplitter, CodeSplitterError, MarkdownSplitter, TextSplitter,
+    Characters, ChunkCapacity, ChunkCapacityError, ChunkConfig, ChunkConfigError, ChunkSizer,
+    CodeSplitter, CodeSplitterError, MarkdownSplitter, TextSplitter,
 };
-use tiktoken_rs::{get_bpe_from_model, CoreBPE};
+use tiktoken_rs::get_bpe_from_model;
 use tokenizers::Tokenizer;
 use tree_sitter::{ffi::TSLanguage, Language};
 
@@ -88,16 +87,13 @@ struct CustomCallback(PyObject);
 
 impl ChunkSizer for CustomCallback {
     /// Determine the size of a given chunk to use for validation
-    fn chunk_size(&self, chunk: &str, capacity: &ChunkCapacity) -> ChunkSize {
+    fn size(&self, chunk: &str) -> usize {
         Python::with_gil(|py| {
-            let size = self
-                .0
+            self.0
                 .call_bound(py, (chunk,), None)
                 .unwrap()
                 .extract::<usize>(py)
-                .unwrap();
-
-            ChunkSize::from_size(size, capacity)
+                .unwrap()
         })
     }
 }
@@ -130,39 +126,18 @@ impl<'text> ByteToCharOffsetTracker<'text> {
     }
 }
 
-#[allow(clippy::large_enum_variant)]
-enum TextSplitterOptions {
-    Characters(TextSplitter<Characters>),
-    CustomCallback(TextSplitter<CustomCallback>),
-    Huggingface(TextSplitter<Tokenizer>),
-    Tiktoken(TextSplitter<CoreBPE>),
+/// Allows for dynamically choosing between different chunk sizers
+struct Sizer(Box<dyn ChunkSizer + 'static + Send>);
+
+impl Sizer {
+    fn new(sizer: impl ChunkSizer + 'static + Send) -> Self {
+        Self(Box::new(sizer))
+    }
 }
 
-impl TextSplitterOptions {
-    #[auto_enum(Iterator)]
-    fn chunks<'splitter, 'text: 'splitter>(
-        &'splitter self,
-        text: &'text str,
-    ) -> impl Iterator<Item = &'text str> + 'splitter {
-        match self {
-            Self::Characters(splitter) => splitter.chunks(text),
-            Self::CustomCallback(splitter) => splitter.chunks(text),
-            Self::Huggingface(splitter) => splitter.chunks(text),
-            Self::Tiktoken(splitter) => splitter.chunks(text),
-        }
-    }
-
-    #[auto_enum(Iterator)]
-    fn chunk_indices<'splitter, 'text: 'splitter>(
-        &'splitter self,
-        text: &'text str,
-    ) -> impl Iterator<Item = (usize, &'text str)> + 'splitter {
-        match self {
-            Self::Characters(splitter) => splitter.chunk_indices(text),
-            Self::CustomCallback(splitter) => splitter.chunk_indices(text),
-            Self::Huggingface(splitter) => splitter.chunk_indices(text),
-            Self::Tiktoken(splitter) => splitter.chunk_indices(text),
-        }
+impl ChunkSizer for Sizer {
+    fn size(&self, chunk: &str) -> usize {
+        self.0.size(chunk)
     }
 }
 
@@ -253,7 +228,7 @@ Args:
 */
 #[pyclass(frozen, name = "TextSplitter")]
 struct PyTextSplitter {
-    splitter: TextSplitterOptions,
+    splitter: TextSplitter<Sizer>,
 }
 
 #[pymethods]
@@ -262,12 +237,13 @@ impl PyTextSplitter {
     #[pyo3(signature = (capacity, overlap=0, trim=true))]
     fn new(capacity: PyChunkCapacity, overlap: usize, trim: bool) -> PyResult<Self> {
         Ok(Self {
-            splitter: TextSplitterOptions::Characters(TextSplitter::new(
+            splitter: TextSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_trim(trim),
-            )),
+                    .with_trim(trim)
+                    .with_sizer(Sizer::new(Characters)),
+            ),
         })
     }
 
@@ -305,13 +281,13 @@ impl PyTextSplitter {
             Tokenizer::from_str(&json).map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: TextSplitterOptions::Huggingface(TextSplitter::new(
+            splitter: TextSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(tokenizer)
+                    .with_sizer(Sizer::new(tokenizer))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -343,18 +319,18 @@ impl PyTextSplitter {
         overlap: usize,
         trim: bool,
     ) -> PyResult<Self> {
-        let tokenizer = json
+        let tokenizer: Tokenizer = json
             .parse()
             .map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: TextSplitterOptions::Huggingface(TextSplitter::new(
+            splitter: TextSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(tokenizer)
+                    .with_sizer(Sizer::new(tokenizer))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -389,13 +365,13 @@ impl PyTextSplitter {
         let tokenizer =
             Tokenizer::from_file(path).map_err(|e| PyException::new_err(format!("{e}")))?;
         Ok(Self {
-            splitter: TextSplitterOptions::Huggingface(TextSplitter::new(
+            splitter: TextSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(tokenizer)
+                    .with_sizer(Sizer::new(tokenizer))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -430,13 +406,13 @@ impl PyTextSplitter {
             get_bpe_from_model(model).map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: TextSplitterOptions::Tiktoken(TextSplitter::new(
+            splitter: TextSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(tokenizer)
+                    .with_sizer(Sizer::new(tokenizer))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -469,13 +445,13 @@ impl PyTextSplitter {
         trim: bool,
     ) -> PyResult<Self> {
         Ok(Self {
-            splitter: TextSplitterOptions::CustomCallback(TextSplitter::new(
+            splitter: TextSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(CustomCallback(callback))
+                    .with_sizer(Sizer::new(CustomCallback(callback)))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -538,42 +514,6 @@ impl PyTextSplitter {
     }
 }
 
-#[allow(clippy::large_enum_variant)]
-enum MarkdownSplitterOptions {
-    Characters(MarkdownSplitter<Characters>),
-    CustomCallback(MarkdownSplitter<CustomCallback>),
-    Huggingface(MarkdownSplitter<Tokenizer>),
-    Tiktoken(MarkdownSplitter<CoreBPE>),
-}
-
-impl MarkdownSplitterOptions {
-    #[auto_enum(Iterator)]
-    fn chunks<'splitter, 'text: 'splitter>(
-        &'splitter self,
-        text: &'text str,
-    ) -> impl Iterator<Item = &'text str> + 'splitter {
-        match self {
-            Self::Characters(splitter) => splitter.chunks(text),
-            Self::CustomCallback(splitter) => splitter.chunks(text),
-            Self::Huggingface(splitter) => splitter.chunks(text),
-            Self::Tiktoken(splitter) => splitter.chunks(text),
-        }
-    }
-
-    #[auto_enum(Iterator)]
-    fn chunk_indices<'splitter, 'text: 'splitter>(
-        &'splitter self,
-        text: &'text str,
-    ) -> impl Iterator<Item = (usize, &'text str)> + 'splitter {
-        match self {
-            Self::Characters(splitter) => splitter.chunk_indices(text),
-            Self::CustomCallback(splitter) => splitter.chunk_indices(text),
-            Self::Huggingface(splitter) => splitter.chunk_indices(text),
-            Self::Tiktoken(splitter) => splitter.chunk_indices(text),
-        }
-    }
-}
-
 /**
 Markdown splitter. Recursively splits chunks into the largest semantic units that fit within the chunk size. Also will attempt to merge neighboring chunks if they can fit within the given chunk size.
 
@@ -663,7 +603,7 @@ Args:
 */
 #[pyclass(frozen, name = "MarkdownSplitter")]
 struct PyMarkdownSplitter {
-    splitter: MarkdownSplitterOptions,
+    splitter: MarkdownSplitter<Sizer>,
 }
 
 #[pymethods]
@@ -672,12 +612,13 @@ impl PyMarkdownSplitter {
     #[pyo3(signature = (capacity, overlap=0, trim=true))]
     fn new(capacity: PyChunkCapacity, overlap: usize, trim: bool) -> PyResult<Self> {
         Ok(Self {
-            splitter: MarkdownSplitterOptions::Characters(MarkdownSplitter::new(
+            splitter: MarkdownSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
+                    .with_sizer(Sizer::new(Characters))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -715,13 +656,13 @@ impl PyMarkdownSplitter {
             Tokenizer::from_str(&json).map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: MarkdownSplitterOptions::Huggingface(MarkdownSplitter::new(
+            splitter: MarkdownSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(tokenizer)
+                    .with_sizer(Sizer::new(tokenizer))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -753,18 +694,18 @@ impl PyMarkdownSplitter {
         overlap: usize,
         trim: bool,
     ) -> PyResult<Self> {
-        let tokenizer = json
+        let tokenizer: Tokenizer = json
             .parse()
             .map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: MarkdownSplitterOptions::Huggingface(MarkdownSplitter::new(
+            splitter: MarkdownSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(tokenizer)
+                    .with_sizer(Sizer::new(tokenizer))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -799,13 +740,13 @@ impl PyMarkdownSplitter {
         let tokenizer =
             Tokenizer::from_file(path).map_err(|e| PyException::new_err(format!("{e}")))?;
         Ok(Self {
-            splitter: MarkdownSplitterOptions::Huggingface(MarkdownSplitter::new(
+            splitter: MarkdownSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(tokenizer)
+                    .with_sizer(Sizer::new(tokenizer))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -840,13 +781,13 @@ impl PyMarkdownSplitter {
             get_bpe_from_model(model).map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: MarkdownSplitterOptions::Tiktoken(MarkdownSplitter::new(
+            splitter: MarkdownSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(tokenizer)
+                    .with_sizer(Sizer::new(tokenizer))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -879,13 +820,13 @@ impl PyMarkdownSplitter {
         trim: bool,
     ) -> PyResult<Self> {
         Ok(Self {
-            splitter: MarkdownSplitterOptions::CustomCallback(MarkdownSplitter::new(
+            splitter: MarkdownSplitter::new(
                 ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
                     .with_overlap(overlap)
                     .map_err(PyChunkConfigError)?
-                    .with_sizer(CustomCallback(callback))
+                    .with_sizer(Sizer::new(CustomCallback(callback)))
                     .with_trim(trim),
-            )),
+            ),
         })
     }
 
@@ -951,42 +892,6 @@ impl PyMarkdownSplitter {
     }
 }
 
-#[allow(clippy::large_enum_variant)]
-enum CodeSplitterOptions {
-    Characters(CodeSplitter<Characters>),
-    CustomCallback(CodeSplitter<CustomCallback>),
-    Huggingface(CodeSplitter<Tokenizer>),
-    Tiktoken(CodeSplitter<CoreBPE>),
-}
-
-impl CodeSplitterOptions {
-    #[auto_enum(Iterator)]
-    fn chunks<'splitter, 'text: 'splitter>(
-        &'splitter self,
-        text: &'text str,
-    ) -> impl Iterator<Item = &'text str> + 'splitter {
-        match self {
-            Self::Characters(splitter) => splitter.chunks(text),
-            Self::CustomCallback(splitter) => splitter.chunks(text),
-            Self::Huggingface(splitter) => splitter.chunks(text),
-            Self::Tiktoken(splitter) => splitter.chunks(text),
-        }
-    }
-
-    #[auto_enum(Iterator)]
-    fn chunk_indices<'splitter, 'text: 'splitter>(
-        &'splitter self,
-        text: &'text str,
-    ) -> impl Iterator<Item = (usize, &'text str)> + 'splitter {
-        match self {
-            Self::Characters(splitter) => splitter.chunk_indices(text),
-            Self::CustomCallback(splitter) => splitter.chunk_indices(text),
-            Self::Huggingface(splitter) => splitter.chunk_indices(text),
-            Self::Tiktoken(splitter) => splitter.chunk_indices(text),
-        }
-    }
-}
-
 /**
 Code splitter. Recursively splits chunks into the largest semantic units that fit within the chunk size. Also will attempt to merge neighboring chunks if they can fit within the given chunk size.
 
@@ -1088,13 +993,25 @@ Args:
 */
 #[pyclass(frozen, name = "CodeSplitter")]
 struct PyCodeSplitter {
-    splitter: CodeSplitterOptions,
+    splitter: CodeSplitter<Sizer>,
 }
 
 impl PyCodeSplitter {
     /// Converts the output of a Python tree-sitter language object into a `Language` struct.
-    fn load_language(language: &Bound<'_, PyAny>) -> Language {
-        unsafe { Language::from_raw(ffi::PyLong_AsVoidPtr(language.as_ptr()) as *const TSLanguage) }
+    fn load_language(language: &Bound<'_, PyAny>) -> PyResult<Language> {
+        unsafe {
+            if ffi::PyCapsule_CheckExact(language.as_ptr()) > 0 {
+                let pointer = ffi::PyCapsule_GetPointer(
+                    language.as_ptr(),
+                    ffi::PyCapsule_GetName(language.as_ptr()),
+                );
+                Ok(Language::from_raw(pointer.cast::<TSLanguage>()))
+            } else {
+                Err(PyTypeError::new_err(
+                "Expected a pointer for the language. Try calling `language()` on the tree-sitter language.",
+                ))
+            }
+        }
     }
 }
 
@@ -1109,16 +1026,15 @@ impl PyCodeSplitter {
         trim: bool,
     ) -> PyResult<Self> {
         Ok(Self {
-            splitter: CodeSplitterOptions::Characters(
-                CodeSplitter::new(
-                    Self::load_language(language),
-                    ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
-                        .with_overlap(overlap)
-                        .map_err(PyChunkConfigError)?
-                        .with_trim(trim),
-                )
-                .map_err(PyCodeSplitterError)?,
-            ),
+            splitter: CodeSplitter::new(
+                Self::load_language(language)?,
+                ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
+                    .with_overlap(overlap)
+                    .map_err(PyChunkConfigError)?
+                    .with_sizer(Sizer::new(Characters))
+                    .with_trim(trim),
+            )
+            .map_err(PyCodeSplitterError)?,
         })
     }
 
@@ -1159,17 +1075,15 @@ impl PyCodeSplitter {
             Tokenizer::from_str(&json).map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: CodeSplitterOptions::Huggingface(
-                CodeSplitter::new(
-                    Self::load_language(language),
-                    ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
-                        .with_overlap(overlap)
-                        .map_err(PyChunkConfigError)?
-                        .with_sizer(tokenizer)
-                        .with_trim(trim),
-                )
-                .map_err(PyCodeSplitterError)?,
-            ),
+            splitter: CodeSplitter::new(
+                Self::load_language(language)?,
+                ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
+                    .with_overlap(overlap)
+                    .map_err(PyChunkConfigError)?
+                    .with_sizer(Sizer::new(tokenizer))
+                    .with_trim(trim),
+            )
+            .map_err(PyCodeSplitterError)?,
         })
     }
 
@@ -1204,22 +1118,20 @@ impl PyCodeSplitter {
         overlap: usize,
         trim: bool,
     ) -> PyResult<Self> {
-        let tokenizer = json
+        let tokenizer: Tokenizer = json
             .parse()
             .map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: CodeSplitterOptions::Huggingface(
-                CodeSplitter::new(
-                    Self::load_language(language),
-                    ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
-                        .with_overlap(overlap)
-                        .map_err(PyChunkConfigError)?
-                        .with_sizer(tokenizer)
-                        .with_trim(trim),
-                )
-                .map_err(PyCodeSplitterError)?,
-            ),
+            splitter: CodeSplitter::new(
+                Self::load_language(language)?,
+                ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
+                    .with_overlap(overlap)
+                    .map_err(PyChunkConfigError)?
+                    .with_sizer(Sizer::new(tokenizer))
+                    .with_trim(trim),
+            )
+            .map_err(PyCodeSplitterError)?,
         })
     }
 
@@ -1257,17 +1169,15 @@ impl PyCodeSplitter {
         let tokenizer =
             Tokenizer::from_file(path).map_err(|e| PyException::new_err(format!("{e}")))?;
         Ok(Self {
-            splitter: CodeSplitterOptions::Huggingface(
-                CodeSplitter::new(
-                    Self::load_language(language),
-                    ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
-                        .with_overlap(overlap)
-                        .map_err(PyChunkConfigError)?
-                        .with_sizer(tokenizer)
-                        .with_trim(trim),
-                )
-                .map_err(PyCodeSplitterError)?,
-            ),
+            splitter: CodeSplitter::new(
+                Self::load_language(language)?,
+                ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
+                    .with_overlap(overlap)
+                    .map_err(PyChunkConfigError)?
+                    .with_sizer(Sizer::new(tokenizer))
+                    .with_trim(trim),
+            )
+            .map_err(PyCodeSplitterError)?,
         })
     }
 
@@ -1305,17 +1215,15 @@ impl PyCodeSplitter {
             get_bpe_from_model(model).map_err(|e| PyException::new_err(format!("{e}")))?;
 
         Ok(Self {
-            splitter: CodeSplitterOptions::Tiktoken(
-                CodeSplitter::new(
-                    Self::load_language(language),
-                    ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
-                        .with_overlap(overlap)
-                        .map_err(PyChunkConfigError)?
-                        .with_sizer(tokenizer)
-                        .with_trim(trim),
-                )
-                .map_err(PyCodeSplitterError)?,
-            ),
+            splitter: CodeSplitter::new(
+                Self::load_language(language)?,
+                ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
+                    .with_overlap(overlap)
+                    .map_err(PyChunkConfigError)?
+                    .with_sizer(Sizer::new(tokenizer))
+                    .with_trim(trim),
+            )
+            .map_err(PyCodeSplitterError)?,
         })
     }
 
@@ -1351,17 +1259,15 @@ impl PyCodeSplitter {
         trim: bool,
     ) -> PyResult<Self> {
         Ok(Self {
-            splitter: CodeSplitterOptions::CustomCallback(
-                CodeSplitter::new(
-                    Self::load_language(language),
-                    ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
-                        .with_overlap(overlap)
-                        .map_err(PyChunkConfigError)?
-                        .with_sizer(CustomCallback(callback))
-                        .with_trim(trim),
-                )
-                .map_err(PyCodeSplitterError)?,
-            ),
+            splitter: CodeSplitter::new(
+                Self::load_language(language)?,
+                ChunkConfig::new(ChunkCapacity::try_from(capacity)?)
+                    .with_overlap(overlap)
+                    .map_err(PyChunkConfigError)?
+                    .with_sizer(Sizer::new(CustomCallback(callback)))
+                    .with_trim(trim),
+            )
+            .map_err(PyCodeSplitterError)?,
         })
     }
 
@@ -1423,9 +1329,11 @@ impl PyCodeSplitter {
 
 #[doc = include_str!("../README.md")]
 #[pymodule]
-fn semantic_text_splitter(m: &Bound<'_, PyModule>) -> PyResult<()> {
-    m.add_class::<PyTextSplitter>()?;
-    m.add_class::<PyMarkdownSplitter>()?;
-    m.add_class::<PyCodeSplitter>()?;
-    Ok(())
+mod semantic_text_splitter {
+    #[pymodule_export]
+    use super::PyCodeSplitter;
+    #[pymodule_export]
+    use super::PyMarkdownSplitter;
+    #[pymodule_export]
+    use super::PyTextSplitter;
 }
