@@ -130,6 +130,13 @@ impl ChunkCapacity {
             Ordering::Equal
         }
     }
+
+    /// Generates a chunk size object based on the size provided from a sizer
+    /// Calculates and stores whether or not it fits within the capacity
+    #[must_use]
+    fn chunk_size(&self, size: usize) -> ChunkSize {
+        ChunkSize::new(self.fits(size), size)
+    }
 }
 
 impl From<usize> for ChunkCapacity {
@@ -193,46 +200,14 @@ impl From<RangeToInclusive<usize>> for ChunkCapacity {
 pub struct ChunkSize {
     /// Whether or not the entire chunk fits within the `ChunkCapacity`
     fits: Ordering,
-    /// max byte offset of the text that fit within the given `ChunkCapacity`.
-    max_chunk_size_offset: Option<usize>,
     /// Size of the chunk, in units used by the sizer.
     size: usize,
 }
 
 impl ChunkSize {
-    /// Generate a chunk size from a given size. Will not be able to compute the
-    /// max byte offset that fits within the capacity.
     #[must_use]
-    pub fn from_size(size: usize, capacity: &ChunkCapacity) -> Self {
-        Self {
-            fits: capacity.fits(size),
-            max_chunk_size_offset: None,
-            size,
-        }
-    }
-
-    /// Generate a chunk size from an iterator of byte ranges for each encoded
-    /// element in the chunk.
-    pub fn from_offsets(
-        offsets: impl Iterator<Item = Range<usize>>,
-        capacity: &ChunkCapacity,
-    ) -> Self {
-        let mut chunk_size = offsets.fold(
-            Self {
-                fits: Ordering::Less,
-                max_chunk_size_offset: None,
-                size: 0,
-            },
-            |mut acc, range| {
-                acc.size += 1;
-                if acc.size <= capacity.max {
-                    acc.max_chunk_size_offset = Some(range.end);
-                }
-                acc
-            },
-        );
-        chunk_size.fits = capacity.fits(chunk_size.size);
-        chunk_size
+    pub fn new(fits: Ordering, size: usize) -> Self {
+        Self { fits, size }
     }
 
     /// Determine whether the chunk size fits within the capacity or not
@@ -251,7 +226,7 @@ impl ChunkSize {
 /// Determines the size of a given chunk.
 pub trait ChunkSizer {
     /// Determine the size of a given chunk to use for validation
-    fn chunk_size(&self, chunk: &str, capacity: &ChunkCapacity) -> ChunkSize;
+    fn size(&self, chunk: &str) -> usize;
 }
 
 /// Indicates there was an error with the chunk configuration.
@@ -436,18 +411,13 @@ where
 
         *cache
             .entry(offset..(offset + chunk.len()))
-            .or_insert_with(|| self.chunk_config.sizer.chunk_size(chunk, &capacity))
+            .or_insert_with(|| capacity.chunk_size(self.chunk_config.sizer.size(chunk)))
     }
 
     /// Check if the chunk is within the capacity. Chunk should be trimmed if necessary beforehand.
     pub fn check_capacity(&mut self, offset: usize, chunk: &str, is_overlap: bool) -> ChunkSize {
         let (offset, chunk) = self.trim_chunk(offset, chunk);
-        let mut chunk_size = self.chunk_size(offset, chunk, is_overlap);
-
-        if let Some(max_chunk_size_offset) = chunk_size.max_chunk_size_offset.as_mut() {
-            *max_chunk_size_offset += offset;
-        }
-        chunk_size
+        self.chunk_size(offset, chunk, is_overlap)
     }
 
     /// If trim chunks is on, trim the str and adjust the offset
@@ -465,9 +435,9 @@ where
         offset: usize,
         levels_with_first_chunk: impl Iterator<Item = (L, &'text str)>,
     ) -> (Option<L>, Option<usize>) {
-        // If we aren't at the highest semantic level, stop iterating sections that go beyond the range of the next level.
-        let mut max_encoded_offset = None;
         let mut semantic_level = None;
+        let mut max_offset = None;
+        let max_size = self.chunk_config.capacity.max;
 
         // We assume that larger levels are also longer. We can skip lower levels if going to a higher level would result in a shorter text
         let levels_with_first_chunk =
@@ -480,17 +450,21 @@ where
             });
 
         for (level, str) in levels_with_first_chunk {
-            let chunk_size = self.check_capacity(offset, str, false);
-            // If this no longer fits, we use the level we are at.
-            if chunk_size.fits.is_gt() {
-                max_encoded_offset = chunk_size.max_chunk_size_offset;
-                break;
+            // Skip tokenizing levels that we know are too small anyway.
+            let len = str.len();
+            if len > max_size {
+                let chunk_size = self.check_capacity(offset, str, false);
+                // If this no longer fits, we use the level we are at.
+                if chunk_size.fits.is_gt() {
+                    max_offset = Some(offset + len);
+                    break;
+                }
             }
             // Otherwise break up the text with the next level
             semantic_level = Some(level);
         }
 
-        (semantic_level, max_encoded_offset)
+        (semantic_level, max_offset)
     }
 
     /// Clear the cached values. Once we've moved the cursor,
@@ -514,14 +488,23 @@ mod tests {
         let chunk = "12345";
 
         assert_eq!(
-            Characters.chunk_size(chunk, &4.into()).fits,
+            ChunkCapacity::from(4)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Greater
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &5.into()).fits,
+            ChunkCapacity::from(5)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
-        assert_eq!(Characters.chunk_size(chunk, &6.into()).fits, Ordering::Less);
+        assert_eq!(
+            ChunkCapacity::from(6)
+                .chunk_size(Characters.size(chunk))
+                .fits,
+            Ordering::Less
+        );
     }
 
     #[test]
@@ -529,19 +512,27 @@ mod tests {
         let chunk = "12345";
 
         assert_eq!(
-            Characters.chunk_size(chunk, &(0..0).into()).fits,
+            ChunkCapacity::from(0..0)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Greater
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(0..5).into()).fits,
+            ChunkCapacity::from(0..5)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Greater
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(5..6).into()).fits,
+            ChunkCapacity::from(5..6)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(6..100).into()).fits,
+            ChunkCapacity::from(6..100)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Less
         );
     }
@@ -551,15 +542,21 @@ mod tests {
         let chunk = "12345";
 
         assert_eq!(
-            Characters.chunk_size(chunk, &(0..).into()).fits,
+            ChunkCapacity::from(0..)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(5..).into()).fits,
+            ChunkCapacity::from(5..)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(6..).into()).fits,
+            ChunkCapacity::from(6..)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Less
         );
     }
@@ -569,7 +566,9 @@ mod tests {
         let chunk = "12345";
 
         assert_eq!(
-            Characters.chunk_size(chunk, &(..).into()).fits,
+            ChunkCapacity::from(..)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
     }
@@ -579,19 +578,27 @@ mod tests {
         let chunk = "12345";
 
         assert_eq!(
-            Characters.chunk_size(chunk, &(0..=4).into()).fits,
+            ChunkCapacity::from(0..=4)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Greater
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(5..=6).into()).fits,
+            ChunkCapacity::from(5..=6)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(4..=5).into()).fits,
+            ChunkCapacity::from(4..=5)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(6..=100).into()).fits,
+            ChunkCapacity::from(6..=100)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Less
         );
     }
@@ -601,15 +608,21 @@ mod tests {
         let chunk = "12345";
 
         assert_eq!(
-            Characters.chunk_size(chunk, &(..0).into()).fits,
+            ChunkCapacity::from(..0)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Greater
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(..5).into()).fits,
+            ChunkCapacity::from(..5)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Greater
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(..6).into()).fits,
+            ChunkCapacity::from(..6)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
     }
@@ -619,61 +632,25 @@ mod tests {
         let chunk = "12345";
 
         assert_eq!(
-            Characters.chunk_size(chunk, &(..=4).into()).fits,
+            ChunkCapacity::from(..=4)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Greater
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(..=5).into()).fits,
+            ChunkCapacity::from(..=5)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
         assert_eq!(
-            Characters.chunk_size(chunk, &(..=6).into()).fits,
+            ChunkCapacity::from(..=6)
+                .chunk_size(Characters.size(chunk))
+                .fits,
             Ordering::Equal
         );
     }
 
-    #[test]
-    fn chunk_size_from_offsets() {
-        let offsets = [0..1, 1..2, 2..3];
-        let chunk_size = ChunkSize::from_offsets(offsets.clone().into_iter(), &1.into());
-        assert_eq!(
-            ChunkSize {
-                fits: Ordering::Greater,
-                size: offsets.len(),
-                max_chunk_size_offset: Some(1)
-            },
-            chunk_size
-        );
-    }
-
-    #[test]
-    fn chunk_size_from_empty_offsets() {
-        let offsets = [];
-        let chunk_size = ChunkSize::from_offsets(offsets.clone().into_iter(), &1.into());
-        assert_eq!(
-            ChunkSize {
-                fits: Ordering::Less,
-                size: offsets.len(),
-                max_chunk_size_offset: None
-            },
-            chunk_size
-        );
-    }
-
-    #[test]
-    fn chunk_size_from_small_offsets() {
-        let offsets = [0..1, 1..2, 2..3];
-        let chunk_size = ChunkSize::from_offsets(offsets.clone().into_iter(), &4.into());
-        assert_eq!(
-            ChunkSize {
-                fits: Ordering::Less,
-                size: offsets.len(),
-                max_chunk_size_offset: Some(3)
-            },
-            chunk_size
-        );
-    }
-
     #[derive(Default)]
     struct CountingSizer {
         calls: AtomicUsize,
@@ -681,9 +658,9 @@ mod tests {
 
     impl ChunkSizer for CountingSizer {
         // Return character version, but count calls
-        fn chunk_size(&self, chunk: &str, capacity: &ChunkCapacity) -> ChunkSize {
+        fn size(&self, chunk: &str) -> usize {
             self.calls.fetch_add(1, atomic::Ordering::SeqCst);
-            Characters.chunk_size(chunk, capacity)
+            Characters.size(chunk)
         }
     }
 
@@ -745,19 +722,6 @@ mod tests {
         );
     }
 
-    #[test]
-    fn test_chunk_size_from_size() {
-        let chunk_size = ChunkSize::from_size(10, &10.into());
-        assert_eq!(
-            ChunkSize {
-                fits: Ordering::Equal,
-                size: 10,
-                max_chunk_size_offset: None
-            },
-            chunk_size
-        );
-    }
-
     #[test]
     fn basic_chunk_config() {
         let config = ChunkConfig::new(10);
@@ -778,7 +742,7 @@ mod tests {
         struct BasicSizer;
 
         impl ChunkSizer for BasicSizer {
-            fn chunk_size(&self, _chunk: &str, _capacity: &ChunkCapacity) -> ChunkSize {
+            fn size(&self, _chunk: &str) -> usize {
                 unimplemented!()
             }
         }
