@@ -5,7 +5,11 @@ use itertools::Itertools;
 use strum::IntoEnumIterator;
 
 use self::fallback::FallbackLevel;
-use crate::{chunk_size::MemoizedChunkSizer, trim::Trim, ChunkConfig, ChunkSizer};
+use crate::{
+    chunk_size::{ChunkSize, MemoizedChunkSizer},
+    trim::Trim,
+    ChunkConfig, ChunkSizer,
+};
 
 #[cfg(feature = "code")]
 mod code;
@@ -125,6 +129,9 @@ struct SemanticSplitRanges<Level>
 where
     Level: SemanticLevel,
 {
+    /// Current cursor in the ranges list, so that we can skip over items we've
+    /// already processed.
+    cursor: usize,
     /// Range of each semantic item and its precalculated semantic level
     ranges: Vec<(Level, Range<usize>)>,
 }
@@ -138,7 +145,7 @@ where
         ranges.sort_unstable_by(|(_, a), (_, b)| {
             a.start.cmp(&b.start).then_with(|| b.end.cmp(&a.end))
         });
-        Self { ranges }
+        Self { cursor: 0, ranges }
     }
 
     /// Retrieve ranges for all sections of a given level after an offset
@@ -146,7 +153,7 @@ where
         &self,
         offset: usize,
     ) -> impl Iterator<Item = (Level, Range<usize>)> + '_ {
-        self.ranges
+        self.ranges[self.cursor..]
             .iter()
             .filter(move |(_, sep)| sep.start >= offset)
             .map(|(l, r)| (*l, r.start..r.end))
@@ -212,8 +219,11 @@ where
     }
 
     /// Clear out ranges we have moved past so future iterations are faster
-    fn update_ranges(&mut self, cursor: usize) {
-        self.ranges.retain(|(_, range)| range.start >= cursor);
+    fn update_cursor(&mut self, cursor: usize) {
+        self.cursor += self.ranges[self.cursor..]
+            .iter()
+            .position(|(_, range)| range.start >= cursor)
+            .unwrap_or_else(|| self.ranges.len() - self.cursor);
     }
 }
 
@@ -238,6 +248,8 @@ where
     semantic_split: SemanticSplitRanges<Level>,
     /// Original text to iterate over and generate chunks from
     text: &'text str,
+    /// Average number of sections in a chunk for each level
+    chunk_stats: ChunkStats,
 }
 
 impl<'sizer, 'text: 'sizer, Sizer, Level> TextChunks<'text, 'sizer, Sizer, Level>
@@ -261,6 +273,7 @@ where
             prev_item_end: 0,
             semantic_split: SemanticSplitRanges::new(offsets),
             text,
+            chunk_stats: ChunkStats::new(),
         }
     }
 
@@ -270,25 +283,27 @@ where
     fn next_chunk(&mut self) -> Option<(usize, &'text str)> {
         // Reset caches so we can reuse the memory allocation
         self.chunk_sizer.clear_cache();
-        self.semantic_split.update_ranges(self.cursor);
-        self.update_next_sections();
+        self.semantic_split.update_cursor(self.cursor);
+        let low = self.update_next_sections();
 
-        let (start, end) = self.binary_search_next_chunk()?;
+        let (start, end) = self.binary_search_next_chunk(low)?;
 
         // Optionally move cursor back if overlap is desired
         self.update_cursor(end);
 
         let chunk = self.text.get(start..end)?;
+
+        self.chunk_stats.update_max_chunk_size(end - start);
+
         // Trim whitespace if user requested it
         Some(self.chunk_sizer.trim_chunk(start, chunk))
     }
 
     /// Use binary search to find the next chunk that fits within the chunk size
-    fn binary_search_next_chunk(&mut self) -> Option<(usize, usize)> {
+    fn binary_search_next_chunk(&mut self, mut low: usize) -> Option<(usize, usize)> {
         let start = self.cursor;
         let mut end = self.cursor;
         let mut equals_found = false;
-        let mut low = 0;
         let mut high = self.next_sections.len().saturating_sub(1);
         let mut successful_index = None;
         let mut successful_chunk_size = None;
@@ -411,13 +426,13 @@ where
     /// Find the ideal next sections, breaking it up until we find the largest chunk.
     /// Increasing length of chunk until we find biggest size to minimize validation time
     /// on huge chunks
-    fn update_next_sections(&mut self) {
+    fn update_next_sections(&mut self) -> usize {
         // First thing, clear out the list, but reuse the allocated memory
         self.next_sections.clear();
 
         let remaining_text = self.text.get(self.cursor..).unwrap();
 
-        let (semantic_level, max_encoded_offset) = self.chunk_sizer.find_correct_level(
+        let (semantic_level, mut max_offset) = self.chunk_sizer.find_correct_level(
             self.cursor,
             self.semantic_split
                 .levels_in_remaining_text(self.cursor)
@@ -429,49 +444,111 @@ where
                 }),
         );
 
-        if let Some(semantic_level) = semantic_level {
-            let sections = self
-                .semantic_split
-                .semantic_chunks(self.cursor, remaining_text, semantic_level)
-                // We don't want to return items at this level that go beyond the next highest semantic level, as that is most
-                // likely a meaningful breakpoint we want to preserve. We already know that the next highest doesn't fit anyway,
-                // so we should be safe to break once we reach it.
-                .take_while_inclusive(move |(offset, _)| {
-                    max_encoded_offset.map_or(true, |max| offset <= &max)
-                })
-                .filter(|(_, str)| !str.is_empty());
-            self.next_sections.extend(sections);
+        let sections = if let Some(semantic_level) = semantic_level {
+            Either::Left(self.semantic_split.semantic_chunks(
+                self.cursor,
+                remaining_text,
+                semantic_level,
+            ))
         } else {
-            let (semantic_level, fallback_max_encoded_offset) =
-                self.chunk_sizer.find_correct_level(
-                    self.cursor,
-                    FallbackLevel::iter().filter_map(|level| {
-                        level
-                            .sections(remaining_text)
-                            .next()
-                            .map(|(_, str)| (level, str))
-                    }),
-                );
+            let (semantic_level, fallback_max_offset) = self.chunk_sizer.find_correct_level(
+                self.cursor,
+                FallbackLevel::iter().filter_map(|level| {
+                    level
+                        .sections(remaining_text)
+                        .next()
+                        .map(|(_, str)| (level, str))
+                }),
+            );
 
-            let max_encoded_offset = match (fallback_max_encoded_offset, max_encoded_offset) {
+            max_offset = match (fallback_max_offset, max_offset) {
                 (Some(fallback), Some(max)) => Some(fallback.min(max)),
                 (fallback, max) => fallback.or(max),
             };
 
-            let sections = semantic_level
-                .unwrap_or(FallbackLevel::Char)
-                .sections(remaining_text)
-                .map(|(offset, text)| (self.cursor + offset, text))
-                // We don't want to return items at this level that go beyond the next highest semantic level, as that is most
-                // likely a meaningful breakpoint we want to preserve. We already know that the next highest doesn't fit anyway,
-                // so we should be safe to break once we reach it.
-                .take_while_inclusive(move |(offset, _)| {
-                    max_encoded_offset.map_or(true, |max| offset <= &max)
-                })
-                .filter(|(_, str)| !str.is_empty());
+            let fallback_level = semantic_level.unwrap_or(FallbackLevel::Char);
 
-            self.next_sections.extend(sections);
+            Either::Right(
+                fallback_level
+                    .sections(remaining_text)
+                    .map(|(offset, text)| (self.cursor + offset, text)),
+            )
+        };
+
+        let mut sections = sections
+            .take_while(move |(offset, _)| max_offset.map_or(true, |max| *offset <= max))
+            .filter(|(_, str)| !str.is_empty());
+
+        // Start filling up the next sections. Since calculating the size of the chunk gets more expensive
+        // the farther we go, we conservatively check for a smaller range to do the later binary search in.
+        let mut low = 0;
+        let mut prev_equals: Option<ChunkSize> = None;
+        let max = self.chunk_config.capacity().max();
+        let mut target_offset = self.chunk_stats.max_chunk_size.unwrap_or(max);
+
+        loop {
+            let prev_num = self.next_sections.len();
+            for (offset, str) in sections.by_ref() {
+                self.next_sections.push((offset, str));
+                if offset + str.len() > (self.cursor.saturating_add(target_offset)) {
+                    break;
+                }
+            }
+            let new_num = self.next_sections.len();
+            // If we've iterated through the whole iterator, break here.
+            if new_num - prev_num == 0 {
+                break;
+            }
+
+            // Check if the last item fits
+            if let Some(&(offset, str)) = self.next_sections.last() {
+                let text_end = offset + str.len();
+                if text_end < target_offset {
+                    break;
+                }
+                let chunk_size = self.chunk_sizer.check_capacity(
+                    offset,
+                    self.text.get(self.cursor..text_end).expect("Invalid range"),
+                    false,
+                );
+
+                let fits = chunk_size.fits();
+                if fits.is_le() {
+                    let final_offset = offset + str.len() - self.cursor;
+                    let size = chunk_size.size().max(1);
+                    let diff = (max - size).max(1);
+                    let avg_size = final_offset.div_ceil(size);
+
+                    target_offset = final_offset
+                        .saturating_add(diff.saturating_mul(avg_size))
+                        .saturating_add(final_offset.div_ceil(10));
+                }
+
+                match fits {
+                    Ordering::Less => {
+                        // We know we can go higher
+                        low = new_num.saturating_sub(1);
+                        continue;
+                    }
+                    Ordering::Equal => {
+                        // Don't update low because it could be a range
+                        // If we've seen a previous equals, we can break if the size is bigger already
+                        if let Some(prev) = prev_equals {
+                            if prev.size() < chunk_size.size() {
+                                break;
+                            }
+                        }
+                        prev_equals = Some(chunk_size);
+                        continue;
+                    }
+                    Ordering::Greater => {
+                        break;
+                    }
+                };
+            }
         }
+
+        low
     }
 }
 
@@ -507,14 +584,54 @@ where
     }
 }
 
+/// Keeps track of the average size of chunks as we go
+#[derive(Debug, Default)]
+struct ChunkStats {
+    /// The size of the biggest chunk we've seen, if we have seen at least one
+    max_chunk_size: Option<usize>,
+}
+
+impl ChunkStats {
+    fn new() -> Self {
+        Self::default()
+    }
+
+    /// Update statistics after the chunk has been produced
+    fn update_max_chunk_size(&mut self, size: usize) {
+        self.max_chunk_size = self.max_chunk_size.map(|s| s.max(size)).or(Some(size));
+    }
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
 
     #[test]
-    fn semantic_ranges_are_sorted() {
-        impl SemanticLevel for usize {}
+    fn chunk_stats_empty() {
+        let stats = ChunkStats::new();
+        assert_eq!(stats.max_chunk_size, None);
+    }
 
+    #[test]
+    fn chunk_stats_one() {
+        let mut stats = ChunkStats::new();
+        stats.update_max_chunk_size(10);
+        assert_eq!(stats.max_chunk_size, Some(10));
+    }
+
+    #[test]
+    fn chunk_stats_multiple() {
+        let mut stats = ChunkStats::new();
+        stats.update_max_chunk_size(10);
+        stats.update_max_chunk_size(20);
+        stats.update_max_chunk_size(30);
+        assert_eq!(stats.max_chunk_size, Some(30));
+    }
+
+    impl SemanticLevel for usize {}
+
+    #[test]
+    fn semantic_ranges_are_sorted() {
         let ranges = SemanticSplitRanges::new(vec![(0, 0..1), (1, 0..2), (0, 1..2), (2, 0..4)]);
 
         assert_eq!(
@@ -522,4 +639,16 @@ mod tests {
             vec![(2, 0..4), (1, 0..2), (0, 0..1), (0, 1..2)]
         );
     }
+
+    #[test]
+    fn semantic_ranges_skip_previous_ranges() {
+        let mut ranges = SemanticSplitRanges::new(vec![(0, 0..1), (1, 0..2), (0, 1..2), (2, 0..4)]);
+
+        ranges.update_cursor(1);
+
+        assert_eq!(
+            ranges.ranges_after_offset(0).collect::<Vec<_>>(),
+            vec![(0, 1..2)]
+        );
+    }
 }
